import { Command, File } from "@lucania/toolbox/server";
import { ConsoleColor, Data } from "@lucania/toolbox/shared";
import { join, posix, relative, resolve, sep as separator } from "path";
import { pathToFileURL } from "url";
import { PackageJson } from "./schema/PackageJson.js";
export class Seam {
    static INSTALLATION_DIRECTORY = "installation";
    static PACKAGE_JSON_FILE = "package.json";
    static NODE_MODULES_DIRECTORY = "node_modules";
    static ON_LOAD_CALLBACK_NAME = "onLoad";
    static ON_UNLOAD_CALLBACK_NAME = "onUnload";
    _rootDirectory;
    _pluginsInstallationDirectory;
    _pluginsInstallationNodeModulesDirectory;
    _pluginRegister;
    _specifierMap;
    _loaded;
    _reloading;
    _installationFlags;
    constructor(options) {
        this._rootDirectory = resolve(options.rootDirectory);
        this._pluginsInstallationDirectory = join(this._rootDirectory, Seam.INSTALLATION_DIRECTORY);
        this._pluginsInstallationNodeModulesDirectory = join(this._pluginsInstallationDirectory, Seam.NODE_MODULES_DIRECTORY);
        this._pluginRegister = {};
        this._specifierMap = {};
        this._loaded = new Set();
        this._reloading = new Set();
        this._installationFlags = [
            "--no-progress",
            "--no-audit",
            "--save",
            "--package-lock=false",
            "--omit=dev"
        ];
    }
    async loadPlugins(...names) {
        const loadResults = {};
        for (const name of names) {
            Data.assert(this.isPluginRegistered(name), `${name} is not a registered plugin!`);
            Data.assert(!this.isPluginLoaded(name), `${name} is already loaded!`);
            const registeredPlugin = this._pluginRegister[name];
            if (registeredPlugin.loadDefinition !== undefined) {
                loadResults[name] = await Promise.resolve(registeredPlugin.loadDefinition.call(name));
            }
            this._loaded.add(name);
        }
        return loadResults;
    }
    async unloadPlugin(name) {
        Data.assert(this.isPluginRegistered(name), `${name} is not a registered plugin!`);
        Data.assert(this.isPluginLoaded(name), `${name} is not loaded!`);
        const registeredPlugin = this._pluginRegister[name];
        if (registeredPlugin.unloadDefinition !== undefined) {
            await Promise.resolve(registeredPlugin.unloadDefinition.call(name));
        }
        this._loaded.delete(name);
    }
    async unloadPlugins(...names) {
        await Promise.all(names.map(name => this.unloadPlugin(name)));
    }
    async reloadPlugin(name, reinstall = true) {
        try {
            this._reloading.add(name);
            if (reinstall) {
                if (await this.isPluginInstalled(name)) {
                    await this.uninstallPlugin(name);
                }
                await this.installPlugins(name);
                await this.registerPlugins(name);
                await this.loadPlugins(name);
            }
            else {
                if (this.isPluginLoaded(name)) {
                    await this.unloadPlugin(name);
                }
                await this.loadPlugins(name);
            }
        }
        finally {
            this._reloading.delete(name);
        }
    }
    async registerPlugins(...names) {
        for (const name of names) {
            let resolution;
            const parent = this.getPackageJsonPath();
            try {
                resolution = import.meta.resolve(name, pathToFileURL(parent));
            }
            catch (error) {
                Data.assert(error instanceof Error, "Threw an error that wasn't of Error type.");
                console.error(`Failed to resolve plugin "${name}" relative to "${parent}". ${error.message}`);
                if (!process.execArgv.includes("--experimental-import-meta-resolve")) {
                    console.error(`Seam currently requires the "--experimental-import-meta-resolve" the flag to be present to resolve plugins correctly.`);
                }
            }
            if (resolution !== undefined) {
                console.log("Executing registration", resolution);
                const module = await import(resolution);
                let loadDefinition = module[Seam.ON_LOAD_CALLBACK_NAME];
                let unloadDefinition = module[Seam.ON_UNLOAD_CALLBACK_NAME];
                if (typeof loadDefinition !== "function") {
                    loadDefinition = undefined;
                }
                if (typeof unloadDefinition !== "function") {
                    unloadDefinition = undefined;
                }
                const { yellow, reset } = ConsoleColor.Common;
                if (loadDefinition === undefined) {
                    console.warn(`${yellow}${name}${reset} does not export a ${Seam.ON_LOAD_CALLBACK_NAME} function.`);
                }
                this._pluginRegister[name] = { loadDefinition, unloadDefinition };
            }
        }
    }
    async unregisterPlugin(name) {
        if (this.isPluginLoaded(name)) {
            await this.unloadPlugin(name);
        }
        delete this._pluginRegister[name];
    }
    async unregisterPlugins(...names) {
        await Promise.all(names.map(name => this.unregisterPlugin(name)));
    }
    /**
     * Installs a list of plugin specifiers to the internal plugin store.
     *
     * @param specifiers An array of NPM specifiers for plugins to install.
     */
    async installPlugins(...specifiers) {
        let packageJson;
        if (await this.isPackageJsonExistent()) {
            packageJson = await this.getPackageJson();
        }
        else {
            packageJson = await this._generatePackageJson();
        }
        specifiers = await this._getNormalizedPluginSpecifiers(specifiers);
        try {
            await Command.execute(`npm install ${specifiers.join(" ")} ${this._installationFlags.join(" ")}`, { currentWorkingDirectory: this._pluginsInstallationDirectory });
            packageJson = await this.getPackageJson();
        }
        catch (error) {
            if (error instanceof Error) {
                let formattedErrorMessage = error.message;
                formattedErrorMessage = formattedErrorMessage.replaceAll(/^/gm, "\t");
                formattedErrorMessage = formattedErrorMessage.replaceAll(/npm ERR! /gm, "");
                const { gray, reset } = ConsoleColor.Common;
                console.error(`NPM failed to install plugins.\n` +
                    `\t${gray}Root: ${this._pluginsInstallationDirectory}\n` +
                    `${formattedErrorMessage}${reset}`);
            }
            return new Array(specifiers.length).fill(undefined);
        }
        return specifiers.map((specifier) => {
            const pluginName = this._findPluginName(specifier, packageJson);
            if (pluginName === undefined) {
                const { yellow, reset } = ConsoleColor.Common;
                console.warn(`Failed to find name of plugin specified by "${yellow}${specifier}${reset}".`);
                console.warn("Checking", packageJson);
            }
            return pluginName;
        });
    }
    async _resolveRelativeFileDependencies(packageJsonPath, relativeTo) {
        console.log(packageJsonPath, relativeTo);
    }
    /**
     * Uninstalls a plugin from the internal plugin store.
     * @param name The name of the plugin to uninstall.
     */
    async uninstallPlugin(name) {
        Data.assert(await this.isPluginInstalled(name), `${name} is not an installed plugin.`);
        await this.unregisterPlugin(name);
        await File.remove(this.getPluginInstallationModuleDirectory(name));
    }
    async uninstallPlugins(...names) {
        await Promise.all(names.map((name) => this.uninstallPlugin(name)));
    }
    async uninstallAllPlugins() {
        for (const name of this.getRegisteredPlugins()) {
            this.unregisterPlugin(name);
        }
        await File.remove(this._pluginsInstallationDirectory);
    }
    isPluginLoaded(name) {
        return this._loaded.has(name);
    }
    isPluginRegistered(name) {
        return name in this._pluginRegister;
    }
    async isPluginInstalled(name) {
        return await File.exists(this.getPluginInstallationModuleDirectory(name));
    }
    getLoadedPlugins() {
        return [...this._loaded];
    }
    getRegisteredPlugins() {
        return Object.keys(this._pluginRegister);
    }
    async _getNormalizedPluginSpecifiers(pluginSpecifiers) {
        const normalizedPluginSpecifiers = [];
        const fileSpecifierPrefix = "file:";
        for (const pluginSpecifier of pluginSpecifiers) {
            if (pluginSpecifier.startsWith(fileSpecifierPrefix)) {
                const filePath = resolve(pluginSpecifier.substring(fileSpecifierPrefix.length));
                const relativeFilePath = relative(this._pluginsInstallationDirectory, filePath);
                const npmRelativeFilePath = posix.join(...relativeFilePath.split(separator));
                normalizedPluginSpecifiers.push(fileSpecifierPrefix + npmRelativeFilePath);
            }
            else if (pluginSpecifier.startsWith(".")) {
                const filePath = resolve(pluginSpecifier);
                const relativeFilePath = relative(this._pluginsInstallationDirectory, filePath);
                const npmRelativeFilePath = posix.join(...relativeFilePath.split(separator));
                normalizedPluginSpecifiers.push(fileSpecifierPrefix + npmRelativeFilePath);
            }
            else if (await File.exists(pluginSpecifier)) {
                const filePath = resolve(pluginSpecifier);
                const relativeFilePath = relative(this._pluginsInstallationDirectory, filePath);
                const npmRelativeFilePath = posix.join(...relativeFilePath.split(separator));
                normalizedPluginSpecifiers.push(fileSpecifierPrefix + npmRelativeFilePath);
            }
            else {
                normalizedPluginSpecifiers.push(pluginSpecifier);
            }
        }
        return normalizedPluginSpecifiers;
    }
    getPluginInstallationDirectory(name) {
        return join(this.getNodeModulesDirectory(), name);
    }
    getPluginPackageJsonPath(name) {
        return join(this.getPluginInstallationDirectory(name), Seam.PACKAGE_JSON_FILE);
    }
    async getPluginPackageJson(name) {
        const packageJsonString = await File.read(this.getPluginPackageJsonPath(name), "utf8");
        const packageJsonData = JSON.parse(packageJsonString);
        return PackageJson.Schema.validate(packageJsonData);
    }
    getPackageJsonPath() {
        return join(this._pluginsInstallationDirectory, Seam.PACKAGE_JSON_FILE);
    }
    async getPackageJson() {
        const packageJsonString = await File.read(this.getPackageJsonPath(), "utf8");
        const packageJsonData = JSON.parse(packageJsonString);
        return PackageJson.Schema.validate(packageJsonData);
    }
    getNodeModulesDirectory() {
        return join(this._pluginsInstallationDirectory, Seam.NODE_MODULES_DIRECTORY);
    }
    async isPackageJsonExistent() {
        return await File.exists(this.getPackageJsonPath());
    }
    getPluginInstallationModuleDirectory(name) {
        return join(this._pluginsInstallationNodeModulesDirectory, name);
    }
    async _generatePackageJson() {
        const packageJson = PackageJson.Schema.validate({ type: "module", name: "@lucania/seam.plugins" });
        await File.write(this.getPackageJsonPath(), JSON.stringify(packageJson, undefined, "    "), "utf8");
        return packageJson;
    }
    _findPluginName(specifier, packageJson) {
        if (specifier in this._specifierMap) {
            return this._specifierMap[specifier];
        }
        if (packageJson.dependencies !== undefined) {
            for (const pluginName in packageJson.dependencies) {
                const dependencySpecifier = packageJson.dependencies[pluginName];
                if (specifier === dependencySpecifier) {
                    this._specifierMap[specifier] = pluginName;
                    this._specifierMap[pluginName] = pluginName;
                    return pluginName;
                }
            }
        }
        return undefined;
    }
}
